import os
import glob
import random
import string
import subprocess

from yaml import safe_load

from charmhelpers import fetch
from charmhelpers.core import host, hookenv
from charmhelpers.contrib.charmsupport.nrpe import NRPE
import wp_helpers


def basenode(service_name):
    if hookenv.hook_name() == 'install':
        hookenv.log('Runnig pre-install hooks')
        for f in glob.glob('exec.d/*/charm-pre-install'):
            if os.path.isfile(f) and os.access(f, os.X_OK):
                subprocess.check_call(['sh', '-c', f])


def unpack_code(service_name):
    config = hookenv.config()
    if not (hookenv.hook_name() in ('install', 'upgrade-charm')
            or config.changed('code_location')):
        return
    host.mkdir(config['install_path'])
    upstream_code = fetch.install_from_config('code_location')
    host.rsync(upstream_code + '/',
               config['install_path'],
               options=['--executability'])  # Because we don't want --delete
    host.rsync('files/wp-upgrade-check.php',
               os.path.join(config['install_path'], 'wp-upgrade-check.php'))
    host.mkdir('{}/wp-content/uploads'.format(config['install_path']),
               owner='www-data', perms=0o755)
    if wp_helpers.wordpress_configured():
        hookenv.log('Running DB upgrade')
        call_php_helper('_db_upgrade.php')


def install_packages(service_name):
    release = host.lsb_release()['DISTRIB_CODENAME']
    if release == 'trusty':
        packages = ['php5-cli', 'php5-mysql', 'php-symfony-yaml', 'php5-curl']
    else:
        packages = ['php-cli', 'php-mysql', 'php-symfony-yaml', 'php-curl', 'php-xml']
    fetch.apt_update()
    fetch.apt_install(packages)


def install_php_memcache(service_name):
    release = host.lsb_release()['DISTRIB_CODENAME']
    if release == 'trusty':
        packages = ['php5-memcached']
    else:
        packages = ['php-memcached']
    fetch.apt_update()
    fetch.apt_install(packages)


def call_php_helper(helper, stdin='', *args):
    config = hookenv.config()
    path = os.path.join(hookenv.charm_dir(), 'files', helper)
    cmd = ['php', path]
    cmd.extend([str(arg) for arg in args])
    hookenv.log(cmd)
    process = subprocess.Popen(
        cmd,
        cwd=config['install_path'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    return process.communicate(stdin)[0]  # spit back stdout+stderr combined


def enable_plugin(*plugins):
    hookenv.log("Enabling plugins: {}".format(plugins))
    hookenv.log(call_php_helper('_enable_plugin.php', '', *plugins))
    # Tell all plugins they may have newly-satisfied dependencies
    plugin_list = call_php_helper('_list_active_plugins.php')
    if plugin_list:
        active_plugins = [x.partition('/')[0] for x in plugin_list.split()]
        for rel_id in hookenv.relation_ids('wordpress-plugin'):
            hookenv.log("{} => {}".format(active_plugins, rel_id))
            hookenv.relation_set(relation_id=rel_id,
                                 active_plugins=' '.join(active_plugins))


def get_option(key):
    release = host.lsb_release()['DISTRIB_RELEASE']
    if release < "16.04":
        script = '_get_option.php'
    else:
        script = '_get_option.php7.php'
    value = call_php_helper(script, '', key)
    return safe_load(value)


def add_option(key, value):
    # Ensure we don't overwrite settings
    if not get_option(key):
        hookenv.log('Adding option: {}'.format(key))
        call_php_helper('_add_option.php', '', key, value)
    else:
        hookenv.log('Option "{}" already in place, skipping.'.format(key))


def enable_akismet(service_name):
    if not wp_helpers.wordpress_configured():
        hookenv.log('Wordpress not configured yet, skipping akismet setup')
        return
    config = hookenv.config()
    hookenv.log('Enabling and configuring akismet plugin...')
    enable_plugin('akismet/akismet.php')
    add_option('akismet_strictness', '0')
    add_option('akismet_show_user_comments_approved', '0')
    add_option('wordpress_api_key', config['akismet_key'])


def enable_plugins(service_name):
    release = host.lsb_release()['DISTRIB_RELEASE']
    hookenv.log("PLUGIN TIME")
    if not wp_helpers.wordpress_configured():
        hookenv.log('Wordpress not configured yet, skipping plugin setup')
        return
    plugins = call_php_helper('_list_inactive_plugins.php').split()
    plugins_to_enable = set()
    plugin_relation = wp_helpers.PluginRelation()
    for relation in plugin_relation['wordpress-plugin']:
        if 'plugin_name' in relation:
            matching_plugins = (x for x in plugins if
                                x.startswith(relation['plugin_name']))
            plugins_to_enable.update(matching_plugins)
        if 'forced_options' in relation:
            if release < "16.04":
                script = '_set_option.php'
            else:
                script = '_set_option.php7.php'
            # The _set_option.php will accept yaml on stdin and iterate.
            hookenv.log(call_php_helper(script,
                                        stdin=relation['forced_options']))
    enable_plugin(*plugins_to_enable)


def first_install(service_name):
    """Perform initial configuratin of wordpress if needed."""
    config = hookenv.config()
    if wp_helpers.wordpress_configured() or not config['initial_settings']:
        hookenv.log('No initial_setting provided or wordpress already '
                    'configured. Skipping first install.')
        return
    hookenv.log('Starting wordpress initial configuration')
    payload = {
        'admin_password': mkpasswd(24),
        'blog_public': 'checked',
        'Submit': 'submit',
    }
    payload.update(safe_load(config['initial_settings']))
    payload['admin_password2'] = payload['admin_password']
    if not payload['blog_public']:
        payload['blog_public'] = 'unchecked'
    required_config = set(('user_name', 'admin_email'))
    missing = required_config.difference(payload.viewkeys())
    if missing:
        hookenv.log('Error: missing wordpress settings: {}'.format(missing))
        return
    wp_helpers.call_wordpress('/wp-admin/install.php?step=2',
                              payload=payload,
                              ssl=config['ssl_enabled'],
                              verify_ssl=False)  # If we use selfsigned cert
    host.write_file(os.path.join(hookenv.charm_dir(), 'initial.passwd'),
                    payload['admin_password'], perms=0o400)


def mkpasswd(length=64):
    return ''.join(random.choice(
                   string.letters + string.digits) for _ in range(length))


def generate_secrets():
    return {k: mkpasswd() for k in ('AUTH_KEY', 'SECURE_AUTH_KEY',
                                    'LOGGED_IN_KEY', 'NONCE_KEY',
                                    'AUTH_SALT', 'SECURE_AUTH_SALT',
                                    'LOGGED_IN_SALT', 'NONCE_SALT')}


def wordpress_static(service_name):
    config = hookenv.config()
    # Wipe wp-overrides.php
    host.write_file(os.path.join(config['install_path'],
                    'wp-overrides.php'), ''),
    host.rsync('files/wp-config.php',
               os.path.join(config['install_path'], 'wp-config.php'))
    host.rsync('files/htaccess',
               os.path.join(config['install_path'], '.htaccess'))


def wipe_secrets(service_name):
    config = hookenv.config()
    files = [os.path.join(hookenv.charm_dir(), 'wp-secrets.json'),
             os.path.join(config['install_path'], 'wp-info.php')]
    for f in files:
        if os.path.exists(f):
            os.unlink(f)


def write_nrpe_checks(service_name):
    config = hookenv.config()
    relation = wp_helpers.NEMRelation()['nrpe-external-master'][0]
    nrpe = NRPE(hostname=relation['nagios_hostname'], primary=True)

    nrpe.add_check(
        shortname='wordpress_http',
        description='Check Wordpress HTTP',
        check_cmd='check_http -I localhost -H {} -p {}'.format(
            config['blog_hostname'], config['port_number'])
    )
    if config['ssl_enabled']:
        nrpe.add_check(
            shortname='wordpress_https',
            description='Check Wordpress HTTPS',
            check_cmd='check_http -I localhost -H {} -p {} -S'.format(
                config['blog_hostname'], config['ssl_port_number'])
        )
    nrpe.add_check(
        shortname='wordpress_upgrades',
        description='Check Wordpress core upgrades',
        check_cmd='/usr/bin/php {}/wp-upgrade-check.php'.format(
            config['install_path'])
    )

    plugin_relations = wp_helpers.PluginRelation()['wordpress-plugin']
    for plugin in plugin_relations:
        plugin_name = plugin.get('plugin_name')
        if plugin_name:
            nrpe.add_check(
                shortname='wordpress_plugin_{}'.format(plugin_name),
                description='Check Wordpress Plugin {}'.format(plugin_name),
                check_cmd='check_file_age -w 31536000 -c 315360000 -f '
                          '{}/wp-content/plugins/{}'
                          ''.format(config['install_path'], plugin_name)

            )

    nrpe.write()


def wipe_nrpe_checks(service_name):
    for f in glob.glob('/var/lib/nagios/export/service__*wordpress_http?.cfg'):
        if os.path.isfile(f):
            os.unlink(f)
    for f in glob.glob('/var/lib/nagios/export/service__*wordpress_plugin*.cfg'):
        if os.path.isfile(f):
            os.unlink(f)


def apparmor_dirs(service_name):
    host.mkdir('/etc/apparmor.d/apache2.d', perms=0o755)
    host.mkdir('/etc/apparmor.d/wordpress.d', perms=0o755)
    host.mkdir('/etc/apparmor.d/webserver.d', perms=0o755)


def reload_apparmor(service_name):
    """Reload apparmor if needed.

    There are 3 situations when it may be needed:
    1. On first install. In this case we rely on a webserver
       subordinate to perform restart.
    2. When rules are updated. This happens when install_path or
       blog_hostname option are updated.
    3. When wp-apparmor.j2 template changes. This should only
       even happen in the upgrade-charm hook
    """
    config = hookenv.config()
    if (config.changed('blog_hostname')
            or config.changed('install_path')
            or hookenv.hook_name() == 'upgrade-charm'):
        host.service_reload('apparmor')


def reload_apache(service_name):
    host.service_reload('apache2')


def notify_webservice_relation(service_name):
    """Send data over webservice relation if any relevant config
    options were changed
    """
    config = hookenv.config()
    options = ['port_number', 'ssl_enabled', 'ssl_port_number', 'redirects',
               'install_path', 'blog_hostname', 'additional_hostnames']
    if not (wp_helpers.object_storage()
            or any(config.changed(option) for option in options)):
        return
    for rel_id in hookenv.relation_ids('webservice'):
        services = wp_helpers.WebserviceRelation().provide_data()['services']
        hookenv.relation_set(relation_id=rel_id, services=services)


def notify_website_relation(service_name):
    """Notify website relation if port number changes."""
    config = hookenv.config()
    hook_name = hookenv.hook_name()
    options = ['port_number', 'blog_hostname', 'additional_hostnames']
    if not (wp_helpers.object_storage()
            or any(config.changed(option) for option in options)
            or hook_name == 'website-relation-joined'):
        return
    for rel_id in hookenv.relation_ids('website'):
        website = wp_helpers.WebsiteRelation()._provide_data(redirects=True)
        hookenv.relation_set(relation_id=rel_id, **website)
